""" This module supplies a class to manage all simulations, regardless of the device simulator.

It works together with the DutView class, which is the subclass of all device simulators.
Features:

* Manages all simulations in a unified way, independent of the actual simulation backend.
* Supports to run simulations on multiple cores in parallel
* Supports to run simulations on a remote server (including file up- and download)

Author: Mario Krattenmacher | mario.krattenmacher@semimod.de
Author: Markus Müller | markus.mueller@semimod.de
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import copy
import logging
import time
import subprocess
import itertools
from joblib import Parallel, delayed
from reprint import output
from pathlib import Path, PurePosixPath, PureWindowsPath
import multiprocessing

from DMT.core import Singleton, print_progress_bar
from DMT.config import DATA_CONFIG
from DMT.exceptions import SimulationUnsuccessful, SimulationFail

# import them always -> can become very annoying otherways (if default is False but one dut is remote)
from tempfile import NamedTemporaryFile
from zipfile import ZipFile

try:
    import paramiko
    from scp import SCPClient, SCPException
except ImportError:
    pass


def upload_progress(filename, size, sent):
    """Callback function for Paramiko SCP Client while uploading files."""
    print_progress_bar(sent, size, prefix="Uploading Simulations", suffix=filename, length=50)


class SimCon(object, metaclass=Singleton):
    """Simulation controller class. SINGLETON design pattern.

    Parameters
    ----------
    n_core  :  int
        Number of cores that shall be used for simulations.
    t_max   :  float
        Timeout for simulations. If a simulation runs longer than t_max in seconds, it is killed.

    Attributes
    ----------
    n_core  :  int
        Number of cores that shall be used for simulations.
    t_max   :  float
        Timeout for simulations. If a simulation runs longer than t_max in seconds, it is killed.
    sim_list :  [{'dut': :class:`~DMT.core.dut_view.DutView`, 'sweep': :class:`~DMT.core.sweep.Sweep`}]
        A list of dicts containing the queued simulations. Each dict holds a 'dut' key value pair and a 'sweep' key value pair.

    ssh_client
        Client to execute SSH commands on a remote server.
    scp_client
        Client to transfer files to a remote server via SCP.
    """

    def __init__(self, n_core=None, t_max=30):
        if n_core is None:
            # Use all available threads by default (for best performance)
            self.n_core = multiprocessing.cpu_count()
        else:
            self.n_core = n_core
        self.t_max = t_max
        self.sim_list = []

        ### ssh stuff
        self.ssh_client = None
        self.scp_client = None

    def clear_sim_list(self):
        """Remove everything from the sim_list"""
        self.sim_list = []

    def append_simulation(self, dut=None, sweep=None):
        """Adds DutViews together with Sweeps to the list of simulations sim_list.

        This methods adds each dut with a copy of each sweep to the simulation list.

        Parameters
        ----------
        dut : :class:`~DMT.core.dut_view.DutView` or [:class:`~DMT.core.dut_view.DutView`]
            Objected of a subclass of DutView. This object describes the device to be simulated and specifies the backend.
        sweep : :class:`~DMT.core.sweep.Sweep` or [:class:`~DMT.core.sweep.Sweep`]
            Definition of the sweep to be performed on the DUT according to the Sweep class.
        """
        if not isinstance(dut, list):
            dut = [dut]
        if isinstance(sweep, list):
            sweep = [copy.deepcopy(sweep_a) for sweep_a in sweep]
        else:
            sweep = [copy.deepcopy(sweep)]

        self.sim_list += [
            {"dut": dut_a, "sweep": sweep_a} for dut_a, sweep_a in itertools.product(dut, sweep)
        ]

    def run_and_read(self, force=False, remove_simulations=False, parallel_read=False):
        """Run all queued simulations and load the results into the Duts' databases.

        Parameters
        ----------
        force : bool, optional
            If True, the simulations will be run and saved back. If False, the simulations will only be run if that has not already been done before. This is ensured using the hash system., by default False
        remove_simulations : bool, optional
            If True, the simulation results will be deleted after read in, by default False. Activate to save disk space.
        parallel_read : bool, optional
            If True, the simulation results are read in using joblib parallel, by default False. Is False because some simulators have issues with this...

        Returns
        -------
        boolean
            True, if no simulation failed. This means it is also true if no simulation was run at all.
        boolean
            True, if any simulation was started. False if all simulations were read from hard disk.
        """
        # reduce number of jobs if we only read a very low number of simulations
        n_jobs = self.n_core if len(self.sim_list) > self.n_core else len(self.sim_list)
        if n_jobs == 0:  # sim list is empty
            return True, False  # all sims were successfull, but no simulations were run
        elif not parallel_read:
            n_jobs = 1

        run_sims = False
        if force:
            logging.info("Simulations forced!")
            sims_to_simulate = self.sim_list
            run_sims = True

        with Parallel(n_jobs=n_jobs, verbose=10) as parallel:
            if not force:
                # check which simulations really need to be run
                n_tot = len(self.sim_list)
                if parallel_read:
                    print("Checking which simulations need to be run in parallel:")
                    sims_checked = parallel(
                        delayed(_check_simulation_needed)(i_sim, n_tot, **sim)
                        for i_sim, sim in enumerate(self.sim_list)
                    )
                else:
                    print("Checking which simulations need to be run:")
                    # parallel not working with VAE modelcard currently since get_circuit is monkey patched
                    sims_checked = [
                        _check_simulation_needed(i_sim, n_tot, **sim)
                        for i_sim, sim in enumerate(self.sim_list)
                    ]

                print_progress_bar(n_tot, n_tot, prefix="Finish", length=50)
                print("\n")  # new line after the progress bar

                sims_to_simulate = []
                # add data to the duts and filter simulations to do
                for sim_to_do, sim_checked in zip(
                    self.sim_list, sims_checked
                ):  # as we are keeping the order, we can copy the data over
                    if sim_checked is None:
                        sims_to_simulate.append(sim_to_do)
                    else:
                        sim_to_do["dut"].data.update(sim_checked)

                run_sims = bool(sims_to_simulate)  # will be False if list is empty

            # remote simulations ?
            if any([sim for sim in sims_to_simulate if sim["dut"].simulate_on_server]):
                self.create_ssh_client()

            # start the simulations using the simulation control.
            process_finished = self.run_simulations(sims_to_simulate)

            if process_finished:
                if parallel_read:
                    print("Reading in the results in parallel:")
                    sims_read = parallel(
                        delayed(_read_process_results)(
                            process["success"], process["dut"], process["sweep"]
                        )
                        for process in process_finished
                    )
                else:
                    print("Reading in the results:")
                    # parallel not working with VAE modelcard currently since get_circuit is monkey patched
                    sims_read = [
                        _read_process_results(process["success"], process["dut"], process["sweep"])
                        for process in process_finished
                    ]
                all_sim_success = all(sim["success"] for sim in sims_read)
                # read data
                for sim in sims_read:
                    # find dut in self.sim_list
                    dut = next(
                        sim_to_do["dut"]
                        for sim_to_do in self.sim_list
                        if sim_to_do["dut"].get_hash() == sim["dut_hash"]
                    )
                    dut.data.update(sim["data"])
            else:
                all_sim_success = True  # no simulations run -> all successfull

        if self.ssh_client is not None:
            self.close_ssh_client()

        if remove_simulations:
            # if storage saving is on, the read simulations can be deleted:
            for sim in self.sim_list:
                sim["dut"].delete_sim_results(sim["sweep"], ignore_errors=True)

        # reset the queue
        self.sim_list = []

        return (
            all_sim_success,
            run_sims,
        )  # the list is empty if no simulations were necessary, empty list -> False

    def create_ssh_client(self):
        """Creates the clients to communicate with the server."""
        self.ssh_client = paramiko.SSHClient()
        self.ssh_client.load_system_host_keys()
        self.ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.ssh_client.connect(
            DATA_CONFIG["server"]["adress"],
            username=DATA_CONFIG["server"]["ssh_user"],
            key_filename=str(Path(DATA_CONFIG["server"]["ssh_key"]).expanduser()),
            disabled_algorithms=dict(pubkeys=["rsa-sha2-512", "rsa-sha2-256"]),
        )

        self.scp_client = SCPClient(
            self.ssh_client.get_transport(), socket_timeout=self.t_max, progress=upload_progress
        )

        # ensure the correct path:
        if DATA_CONFIG["server"]["unix"]:
            DATA_CONFIG["server"]["simulation_path"] = PurePosixPath(
                DATA_CONFIG["server"]["simulation_path"]
            )
        else:
            DATA_CONFIG["server"]["simulation_path"] = PureWindowsPath(
                DATA_CONFIG["server"]["simulation_path"]
            )

        # make sure that the target folder exists
        for folder in reversed(DATA_CONFIG["server"]["simulation_path"].parents):
            self.ssh_client.exec_command("mkdir -p " + str(folder))
        self.ssh_client.exec_command("mkdir -p " + str(DATA_CONFIG["server"]["simulation_path"]))

    def close_ssh_client(self):
        """Closes the ssh connection again."""
        self.ssh_client.close()

        self.ssh_client = None
        self.scp_client = None

    def copy_zip_to_server(self, sims_to_zip):
        """Copies the simulation data to the server. Before doing this, old simulation data is deleted

        Parameters
        ----------
        sims_to_zip : list[dict]
            A list of dictionaries with 2 keys:
            dut : DutView
            sweep : Sweep
        """

        def add_to_zip(folder, rel_to):
            # add folder (not really needed, if there is any file in the folder, but we don't know this....)
            zip_ref.write(folder, arcname=folder.relative_to(rel_to))
            # add all data inside folder
            for child in folder.iterdir():
                if child.is_file():
                    zip_ref.write(child, arcname=child.relative_to(rel_to))
                else:
                    add_to_zip(child, rel_to)

        assert self.ssh_client is not None
        assert self.scp_client is not None

        sim_path_on_server = DATA_CONFIG["server"]["simulation_path"]
        commands = []
        # delete possible old directories:
        for sim_to_zip in sims_to_zip:
            sim_folder = sim_to_zip["dut"].get_sim_folder(sim_to_zip["sweep"])
            dut_folder = sim_folder.parts[-2]
            sweep_folder = sim_folder.parts[-1]
            commands.append("rm -rf " + str(sim_path_on_server / dut_folder / sweep_folder))

        for command in commands:
            self.ssh_client.exec_command(command)

        with NamedTemporaryFile() as path_zip:  # dut.get_sim_folder(sweep).relative_to(dut.sim_dir)
            with ZipFile(path_zip, "w") as zip_ref:
                for sim_to_zip in sims_to_zip:
                    add_to_zip(
                        sim_to_zip["dut"].get_sim_folder(sim_to_zip["sweep"]),
                        sim_to_zip["dut"].sim_dir,
                    )
                    # add "central" VA-Files -.-
                    if not sim_to_zip["dut"]._copy_va_files:
                        va_files_dir = sim_to_zip["dut"].sim_dir / "VA_codes"
                        for vafile in sim_to_zip["dut"]._list_va_file_contents:
                            dir_code = va_files_dir / vafile.get_tree_hash()
                            add_to_zip(
                                dir_code,
                                sim_to_zip["dut"].sim_dir,
                            )

            # transfer and save name
            self.scp_client.put(path_zip.name, remote_path=str(sim_path_on_server))
            name_zip_file = Path(path_zip.name).name

        # unzip
        # here we should wait for finish
        _stdin, stdout, _stderr = self.ssh_client.exec_command(
            "unzip -o -d " + str(sim_path_on_server) + " " + str(sim_path_on_server / name_zip_file)
        )
        stdout.channel.set_combine_stderr(True)
        output = stdout.readlines()

        # delete temp zip on the server
        self.ssh_client.exec_command("rm -f " + str(sim_path_on_server / name_zip_file))

    def copy_from_server(self, dut, sweep, zip_result=True):
        """Collects the simulation data from the server.

        Parameters
        ----------
        dut : DutView
        sweep : Sweep
        zip_result : bool, optional
            If True, the result is zipped before transfer, the zip is copied and then unzipped locally.
        """
        sim_folder = dut.get_sim_folder(sweep)
        root = sim_folder.parent
        sweep_folder = sim_folder.parts[-1]
        dut_folder = sim_folder.parts[-2]
        if zip_result:
            # delete possible old zip:
            self.ssh_client.exec_command(
                "rm -f {0:s}.zip".format(
                    str(DATA_CONFIG["server"]["simulation_path"] / dut_folder / sweep_folder)
                )
            )  # remove to be sure

            # create new zip and copy it via scp
            channel_zip = self.ssh_client.get_transport().open_session(timeout=self.t_max)
            channel_zip.exec_command(
                "cd {0:s} && zip -r {1:s}.zip ./{1:s}".format(
                    str(DATA_CONFIG["server"]["simulation_path"] / dut_folder), sweep_folder
                )
            )
            while not channel_zip.exit_status_ready():
                time.sleep(0.5)

            try:
                self.scp_client.get(
                    str(DATA_CONFIG["server"]["simulation_path"] / dut_folder / sweep_folder)
                    + ".zip",
                    local_path=str(root),
                )
            except (SCPException, paramiko.SSHException, TimeoutError) as err:
                raise FileNotFoundError from err

            path_zip = sim_folder.with_suffix(".zip")
            with ZipFile(path_zip, "r") as zip_ref:
                zip_ref.extractall(root)

            path_zip.unlink()
        else:
            try:
                self.scp_client.get(
                    str(DATA_CONFIG["server"]["simulation_path"] / dut_folder / sweep_folder),
                    local_path=str(root),
                    recursive=True,
                )
            except (SCPException, paramiko.SSHException) as err:
                # reraise it in order to allow run_and_read to go on and try again in 2 seconds
                raise FileNotFoundError from err

    def copy_log_from_server(self, dut, sweep):
        """Collects the simulation log file from the server.

        Parameters
        ----------
        dut : DutView
        sweep : Sweep
        """
        sim_folder = dut.get_sim_folder(sweep)
        root = sim_folder.parent
        sweep_folder = sim_folder.parts[-1]
        dut_folder = sim_folder.parts[-2]
        try:
            self.scp_client.get(
                str(
                    DATA_CONFIG["server"]["simulation_path"] / dut_folder / sweep_folder / "sim.log"
                ),
                local_path=str(root / sweep_folder),
                recursive=True,
            )
        except (SCPException, paramiko.SSHException) as err:
            # reraise it in order to allow run_and_read to go on and try again in 2 seconds
            raise FileNotFoundError from err

    def run_simulations(self, sim_list):
        """Runs all given simulations in parallel.

        Parameters
        ----------
        sim_list :  [{}]
            List of dictionaries, each dictionary has a 'dut': :class:`~DMT.core.DutView` and 'sweep': :class:`~DMT.core.Sweep` key value pair.

        Returns
        -------
        success  :  list[process]
            List of finished processes
        """
        if len(sim_list) == 0:
            return []

        # test if same simulation is added twice.
        set_dut_hashes = set([sim_i["dut"].get_hash() for sim_i in sim_list])

        list_to_delete = []
        for dut_hash in set_dut_hashes:
            list_sweep_hashes = []
            for i_sim, sim_a in enumerate(sim_list):
                if sim_a["dut"].get_hash() == dut_hash:
                    if sim_a["sweep"].get_hash() in list_sweep_hashes:
                        list_to_delete.append(i_sim)
                    else:
                        list_sweep_hashes.append(sim_a["sweep"].get_hash())

        for to_delete in sorted(list_to_delete, reverse=True):
            del sim_list[to_delete]

        # start simulations
        process_running = []
        process_finished = []
        finished = False
        n = 0
        n_total = len(sim_list)

        # prepare simulations
        print_progress_bar(0, len(sim_list), prefix="Preparing Simulations", length=50)
        sims_to_zip = []

        # if True: use pbs job scheduler
        pbs = DATA_CONFIG["server"]["use_pbs"] and DATA_CONFIG["backend_remote"]
        if DATA_CONFIG["progress_minimal"]:
            len_output = 2
        else:
            len_output = self.n_core + 7

        for i_sim, sim in enumerate(sim_list):
            sweep = sim["sweep"]
            dut = sim["dut"]
            print_progress_bar(i_sim, len(sim_list), prefix="Preparing Simulations", length=50)
            dut.prepare_simulation(sweep)
            if dut.simulate_on_server:
                sims_to_zip.append({"dut": dut, "sweep": sweep})

        print_progress_bar(len(sim_list), len(sim_list), prefix="Preparing Simulations", length=50)
        print("\n")  # new line after the progress bar

        if sims_to_zip:
            print("Uploading simulation input files and folders to server...")
            self.copy_zip_to_server(sims_to_zip)
            print("finish uploading.")

        # do not print download status
        if self.scp_client is not None:
            self.scp_client._progress = False

        with output(output_type="list", initial_len=len_output, interval=0) as output_list:
            while not finished:
                # run infinite processes parallel on the server
                # if (len([process for process in process_running if not process['backend_remote']]) < self.n_core) and (len(sim_list) > 0 ):
                if (len([process for process in process_running]) < self.n_core) and (
                    len(sim_list) > 0
                ):
                    # take the next element from the self.sim_list and start it
                    sim = sim_list[0]
                    sim_list = sim_list[1:]
                    # start the simulation on this core
                    sweep = sim["sweep"]
                    dut = sim["dut"]
                    if (
                        not hasattr(dut, "t_max") or dut.t_max is None
                    ):  # make sure t_max is set in every simulated dut
                        dut.t_max = self.t_max

                    if dut.simulate_on_server:
                        pid = self.run_simulation_remote(dut, sweep, pbs=pbs)
                        process = 0

                    else:
                        process = self.run_simulation_local(dut, sweep)
                        pid = process.pid

                    if pid == 0:
                        continue  # failed to start simulation, just wait and try again

                    if hasattr(dut, "zip_result"):
                        zip_result = dut.zip_result
                    else:
                        zip_result = True  # per default True, it is better because scp struggles with many files...

                    n += 1
                    t0 = time.time()
                    process_running.append(
                        {
                            "n": n,
                            "t0": t0,
                            "dt": t0,
                            "dut": dut,
                            "sweep": sweep,
                            "process": process,
                            "pid": pid,
                            "success": True,
                            "backend_remote": dut.simulate_on_server,
                            "last_poll": 0,
                            "zip_result": zip_result,
                        }
                    )

                # check for finished processes. DO THIS BEFORE TIMEOUT CHECKING.
                for p in process_running:
                    process = p["process"]
                    if p["backend_remote"]:
                        p["last_poll"] += 1
                        if (
                            p["last_poll"] % 5 == 0
                        ):  # every 20th round -> every 2 seconds (is this too much?)
                            if pbs:
                                # use qstat
                                _stdin, stdout, _stderr = self.ssh_client.exec_command(
                                    ("qstat_script " + str(p["pid"]))
                                )
                                out = str(stdout.read())
                                if (
                                    "Unknown Job" in out or out == "b''"
                                ):  # if job finished, these strings are returned
                                    try:
                                        self.copy_from_server(
                                            p["dut"], p["sweep"], zip_result=p["zip_result"]
                                        )
                                        process_finished.append(p)
                                        try:
                                            p["dut"].validate_simulation_successful(p["sweep"])
                                        except (
                                            SimulationFail,
                                            SimulationUnsuccessful,
                                            FileNotFoundError,
                                        ):
                                            p["success"] = False
                                    except (SimulationUnsuccessful, FileNotFoundError):
                                        pass  # just try again

                            else:  # copy everything and check => slow
                                try:
                                    self.copy_log_from_server(p["dut"], p["sweep"])
                                    p["dut"].validate_simulation_successful(p["sweep"])
                                    self.copy_from_server(
                                        p["dut"], p["sweep"], zip_result=p["zip_result"]
                                    )
                                    process_finished.append(p)
                                except (SimulationUnsuccessful, FileNotFoundError):
                                    pass
                                except SimulationFail:
                                    p["success"] = False
                                    process_finished.append(p)

                    else:
                        returncode = process.poll()
                        if returncode is not None:
                            if (
                                returncode != 0
                                and returncode != 134
                                and returncode != 139
                                and returncode != 1
                            ):  # 134 sometimes happens but still ads works...
                                p["success"] = False

                            process_finished.append(p)

                # check for timeouts
                t = time.time()
                for p in process_running:
                    p["dt"] = t - p["t0"]
                    if (p["dt"] > p["dut"].t_max) and (
                        p["dt"] > self.t_max
                    ):  # both t_max have to be smaller than the simulation time
                        if not p["backend_remote"]:
                            p["process"].kill()
                        # TODO: kill with pbs
                        p["success"] = False
                        process_finished.append(p)

                # remove finished processes from running processes
                for p in process_finished:
                    if p in process_running:
                        process_running.remove(p)

                # update status that is displayed on the console
                len_progress = 20  # number of #
                progress = int(
                    len(process_finished)
                    / (len(sim_list) + len(process_running) + len(process_finished))
                    * len_progress
                )
                output_list[0] = "DMT is now simulating!         "
                output_list[1] = (
                    "finished: "
                    + str(len(process_finished))
                    + " of "
                    + str(n_total)
                    + ":["
                    + "#" * progress
                    + " " * (len_progress - progress)
                    + "]"
                )

                if not DATA_CONFIG["progress_minimal"]:
                    output_list[2] = "-------------------------------"
                    output_list[3] = "| sim_n | pid        | dt     |"
                    output_list[4] = "-------------------------------"
                    for i in range(self.n_core):
                        try:
                            p = process_running[i]
                            str_ = "|{:^7d}|{:^12d}|{:^8.1f}|".format(p["n"], p["pid"], p["dt"])
                        except (KeyError, IndexError):
                            str_ = "|{:^7s}|{:^12s}|{:^8.1f}|".format("x", "x", 0)

                        output_list[i + 5] = str_

                    output_list[-2] = "-------------------------------"
                    output_list[-1] = "                               "

                # are we finished?
                if len(process_running) == 0 and len(sim_list) == 0:
                    finished = True
                elif len(process_running) == self.n_core or len(sim_list) == 0:
                    time.sleep(0.1)

        # print download status
        if self.scp_client is not None:
            self.scp_client._progress = True

        return process_finished

    def run_simulation_local(self, dut, sweep):
        """Starts the simulation

        Parameters
        ----------
        dut : DutView
        sweep : Sweep
        """
        sim_folder = dut.get_sim_folder(sweep)
        logging.info(
            "Started the simulation for the dut %s of the sweep %s!", dut.get_hash(), sweep.name
        )
        logging.debug("The simulation folder of this simulation is %s", sim_folder)
        log_file = open(sim_folder / "sim.log", "w")
        log_file.write(f"The simulation command is\n{dut.get_start_sim_command()}\n\n")
        return subprocess.Popen(
            dut.get_start_sim_command().split(),
            shell=False,
            cwd=sim_folder,
            stderr=subprocess.STDOUT,
            stdout=log_file,
        )

    def run_simulation_remote(self, dut, sweep, pbs=False):
        """Starts the remote simulation

        Parameters
        ----------
        dut : DutView
        sweep : Sweep
        pbs : Boolean

        Returns
        -------
        pid : int
            0 if failed, -1 if running via ssh directly and id of job for PBS simulation.
        """
        sim_folder = dut.get_sim_folder(sweep)
        sweep_folder = sim_folder.parts[-1]
        dut_folder = sim_folder.parts[-2]

        logging.info(
            "Started the remote simulation for the dut %s of the sweep %s!",
            dut.get_hash(),
            sweep.name,
        )
        logging.debug("The simulation folder of this simulation is %s", sim_folder)

        # start a subprocess with the ssh command
        if not pbs:
            _stdin, _stdout, _stderr = self.ssh_client.exec_command(
                (
                    "cd "
                    + str(DATA_CONFIG["server"]["simulation_path"] / dut_folder / sweep_folder)
                    + ";"
                    + dut.get_start_sim_command()
                    + " > sim.log &"
                )
            )
            ## useful for debugging:
            # for line in iter(_stdout.readline, ""):
            #     print(line, end="")
            # for line in iter(_stderr.readline, ""):
            #     print(line, end="")
            return -1
        else:
            _stdin, stdout, _stderr = self.ssh_client.exec_command(
                (
                    "cd "
                    + str(DATA_CONFIG["server"]["simulation_path"] / dut_folder / sweep_folder)
                    + ";"
                    + DATA_CONFIG["server"]["command_qsub"]
                )
            )
            output = stdout.read()
            _error = _stderr.read()
            id_ = "".join([n for n in str(output).split(".")[0] if n.isdigit()])
            try:
                return int(id_)
            except ValueError:
                return 0


def _check_simulation_needed(i_sim, n_tot, dut=None, sweep=None):
    """Function to check if the simulation is needed or already present in the database

    Parameter
    -----------
    dut : DMT.core.DutView
    sweep : DMT.core.Sweep

    Returns
    -------
    {key: DMT.core.Dataframe}
        In case the data is read from database or previous simulation.
    None
        In case the simulation must be done.
    """
    dut_name = dut.name + str(dut.get_hash())
    sim_name = sweep.name + "_" + sweep.get_hash()
    # print("Check: dut: {:s}, sweep: {:s}".format(dut_name, sim_name))
    print_progress_bar(i_sim, n_tot, prefix="Progress", length=50)
    if dut.check_existence_sweep(sweep):
        logging.info("Simulation of DuT %s with sweep %s loaded from database.", dut_name, sim_name)
    else:
        # if not in dut.data and not in dut.db
        try:
            # was it simulated already successfully ?
            dut.validate_simulation_successful(sweep)
            logging.info(
                "Simulation of DuT %s with sweep %s already done and results are valid, only data needs to be read.",
                dut_name,
                sim_name,
            )
            logging.debug(
                "The simulation folder of this simulation was %s", dut.get_sim_folder(sweep)
            )
            dut.add_data(sweep)
        except SimulationFail:
            print("Simulation of DuT %s with sweep %s already done and failed!", dut_name, sim_name)
        # except (SimulationUnsuccessful, FileNotFoundError, IndexError, struct.error):
        except:  # all exceptions should be re-simulated
            # ok simulate it!
            dut.delete_sim_results(sweep, ignore_errors=True)  # remove for safety
            logging.info("Simulation of DuT %s with sweep %s needed.", dut_name, sim_name)
            return None

    return dut.data


def _read_process_results(success, dut, sweep):
    """Read the process results

    Parameter
    -----------
    success : bool
    dut : DMT.core.DutView
    sweep : DMT.core.Sweep

    Returns
    -------
    {'success': success, 'dut_hash':dut.get_hash(), 'data':dut.data}
    """
    dut_name = dut.name + str(dut.get_hash())
    sim_name = sweep.name + "_" + sweep.get_hash()
    sim_folder = dut.get_sim_folder(sweep)
    print("Read: dut: {:s}, sweep: {:s}".format(dut_name, sim_name))
    # inform data_manager about the finished simulations
    try:
        if success:
            dut.add_data(sweep)
            logging.info("Simulation of DuT %s with sweep %s successfull.", dut_name, sim_name)
        else:
            color_red = "\033[91m"
            color_end = "\033[0m"
            print(
                "{0:s}Simulation of DuT {1:s} with sweep {2:s} failed.{3:s}".format(
                    color_red, dut_name, sim_name, color_end
                )
            )
            print(
                "{0:s}Simulation folder: {1:s} {2:s}".format(color_red, str(sim_folder), color_end)
            )
            print((sim_folder / "sim.log").read_text())
            logging.info("Simulation of DuT %s with sweep %s failed.", dut_name, sim_name)
    except (SimulationUnsuccessful, FileNotFoundError, KeyError):
        color_red = "\033[91m"
        color_end = "\033[0m"
        print(
            "{0:s}Simulation of DuT {1:s} with sweep {2:s} failed.{3:s}".format(
                color_red, dut_name, sim_name, color_end
            )
        )
        print("{0:s}Simulation folder: {1:s} {2:s}".format(color_red, str(sim_folder), color_end))
        print((sim_folder / "sim.log").read_text())
        logging.info("Simulation of DuT %s with sweep %s failed.", dut_name, sim_name)

    return {"success": success, "dut_hash": dut.get_hash(), "data": dut.data}



import types
import numpy as np
from pathlib import Path
from DMT.core import Sweep, SimCon, Plot, specifiers, DutType, MCard
from DMT.core.circuit import Circuit, CircuitElement, HICUML2_HBT, SHORT, VOLTAGE, RESISTANCE
from DMT.xyce import DutXyce
from DMT.ngspice import DutNgspice

# path to DMT test cases
path_test = Path(__file__).resolve().parent.parent.parent.parent / "test"

# load a hicum/L2 modelcard library by loading the corresponding *va code.
modelcard = MCard(
    ["C", "B", "E", "S", "T"],
    default_module_name="",
    default_subckt_name="",
    va_file=path_test / "test_interface_xyce" / "hicuml2v2p4p0_xyce.va",
)
modelcard.load_model_parameters(
    path_test / "test_core_no_interfaces" / "test_modelcards" / "npn_full.lib",
)
modelcard.update_from_vae(remove_old_parameters=True)

# bind the correct get_circuit method in order to make it easily available for simulation
def get_circuit(self):
    """

    Parameter
    ------------
    modelcard : :class:`~DMT.core.MCard`

    Returns
    -------
    circuit : :class:`~DMT.core.circuit.Circuit`

    """
    circuit_elements = []
    # model instance
    circuit_elements.append(
        CircuitElement(
            "hicumL2_test",
            "Q_H",
            [f"n_{node.upper()}" for node in self.nodes_list],
            parameters=self,
        )
    )

    # BASE NODE CONNECTION #############
    # shorts for current measurement
    circuit_elements.append(CircuitElement(SHORT, "I_B", ["n_B", "n_B_FORCED"]))
    # COLLECTOR NODE CONNECTION #############
    circuit_elements.append(CircuitElement(SHORT, "I_C", ["n_C", "n_C_FORCED"]))
    # EMITTER NODE CONNECTION #############
    circuit_elements.append(CircuitElement(SHORT, "I_E", ["n_E", "0"]))
    # add sources and thermal resistance
    circuit_elements.append(
        CircuitElement(
            VOLTAGE, "V_B", ["n_B_FORCED", "0"], parameters=[("Vdc", "V_B"), ("Vac", "V_B_ac")]
        )
    )
    circuit_elements.append(
        CircuitElement(
            VOLTAGE, "V_C", ["n_C_FORCED", "0"], parameters=[("Vdc", "V_C"), ("Vac", "V_C_ac")]
        )
    )
    circuit_elements += ["V_B=0", "V_C=0", "ac_switch=0", "V_B_ac=1-ac_switch", "V_C_ac=ac_switch"]

    return Circuit(circuit_elements)


modelcard.get_circuit = types.MethodType(get_circuit, modelcard)

duts = []
# init an Xyce Dut
duts.append(DutXyce(None, DutType.npn, modelcard, nodes="C,B,E,S,T", reference_node="E"))
# DMT uses the exact same interface for all circuit simulators, e.g. the call for a ngspice simulation would be:
duts.append(DutNgspice(None, DutType.npn, modelcard, nodes="C,B,E,S,T", reference_node="E"))
# isn't it great?!?

# create a sweep (all DMT Duts can use this!)
# some column names we want to simulate and plot
col_ve = specifiers.VOLTAGE + "E"
col_vb = specifiers.VOLTAGE + "B"
col_vc = specifiers.VOLTAGE + "C"
col_vbe = specifiers.VOLTAGE + ["B", "E"]
col_vcb = specifiers.VOLTAGE + ["C", "B"]
col_vbc = specifiers.VOLTAGE + ["B", "C"]
col_ic = specifiers.CURRENT + "C"
col_freq = specifiers.FREQUENCY
col_ft = specifiers.TRANSIT_FREQUENCY
sweepdef = [
    {"var_name": col_freq, "sweep_order": 4, "sweep_type": "LIST", "value_def": [10e9]},
    {"var_name": col_vb, "sweep_order": 3, "sweep_type": "LIN", "value_def": [0.5, 1, 51]},
    {
        "var_name": col_vc,
        "sweep_order": 3,
        "sweep_type": "SYNC",
        "master": col_vb,
        "offset": col_vcb,
    },
    {"var_name": col_vcb, "sweep_order": 2, "sweep_type": "LIST", "value_def": [-0.5, 0, 0.5]},
    {"var_name": col_ve, "sweep_order": 1, "sweep_type": "CON", "value_def": [0]},
]
outputdef = ["I_C", "I_B"]
othervar = {"TEMP": 300}
sweep = Sweep("gummel", sweepdef=sweepdef, outputdef=outputdef, othervar=othervar)

# The simulation controller can control all simulators implemented in DMT!
sim_con = SimCon()

# Add the desired simulation to the queue and start the simulation
sim_con.append_simulation(dut=duts, sweep=sweep)
sim_con.run_and_read(force=True, remove_simulations=False)

# Plot and save as pdf
plt_ic = Plot(
    plot_name="J_C(V_BE)",
    x_specifier=col_vbe,
    y_specifier=col_ic,
    y_scale=1e3,
    y_log=True,
    legend_location="lower right",
)
plt_ft = Plot(
    plot_name="F_T(J_C)",
    x_specifier=col_ic,
    x_scale=1e3,
    x_log=True,
    y_specifier=col_ft,
    legend_location="upper left",
)

for dut in duts:
    name = dut.name.split("_")[0] + " "
    # Read back the iv data of the circuit simulator
    data = dut.get_data(sweep=sweep)
    # Ensure derived quantities, e.g. the circuit simulator only gives you S parameters
    data.ensure_specifier_column(col_vbe, ports=["B", "C"])
    data.ensure_specifier_column(col_vbc, ports=["B", "C"])
    data.ensure_specifier_column(col_ft, ports=["B", "C"])

    for i_vbc, vbc, data_vbc in data.iter_unique_col(col_vbc, decimals=3):
        vbc = np.real(vbc)
        plt_ic.add_data_set(
            data_vbc[col_vbe], data_vbc[col_ic], label=name + col_vbc.to_legend_with_value(vbc)
        )
        plt_ft.add_data_set(
            data_vbc[col_ic], data_vbc[col_ft], label=name + col_vbc.to_legend_with_value(vbc)
        )

plt_ic.x_limits = 0.7, 1
plt_ic.y_limits = 1e-2, 1e2
plt_ft.x_limits = 1e-2, 1e2
plt_ft.y_limits = 0, 420

plt_ic.plot_pyqtgraph(show=False)
plt_ft.plot_pyqtgraph()

# plt_ic.save_tikz(
#     Path(__file__).parent.parent / "_static" / "running_a_simulation",
#     standalone=True,
#     build=True,
#     clean=True,
#     width="3in",
# )
# plt_ft.save_tikz(
#     Path(__file__).parent.parent / "_static" / "running_a_simulation",
#     standalone=True,
#     build=True,
#     clean=True,
#     width="3in",
# )



from pathlib import Path
from DMT.core import read_elpa, read_DEVICE_bin
import os
import time


folder_path = Path(__file__).resolve().parent


def test_read_elpa():
    start_time = time.time()
    _device_elpa = read_elpa(folder_path / "HBT_vbc.elpa")
    end_time = time.time()
    print("read_elpa took ", end_time - start_time, " seconds")


def test_read_DEVICE_internal():
    start_time = time.time()
    _df_internal = read_DEVICE_bin(folder_path / "test_data" / "idc1")
    end_time = time.time()
    print("read_DEVICE_bin took ", end_time - start_time, " seconds")


if __name__ == "__main__":
    test_read_elpa()
    test_read_DEVICE_internal()
    dummy = 1



""" Base class to handle verilog-a modelcards.

Author: Mario Krattenmacher | Mario.Krattenmacher@semimod.de
Author: Markus Müller       | Markus.Mueller3@tu-dresden.de
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
from __future__ import annotations
import numpy as np
import re
import os
import pickle
import logging
import scipy.io as sciio
import ast
import operator
import warnings
from pathlib import Path
from typing import Union, Optional
from types import ModuleType

try:
    from semver.version import Version as VersionInfo
except ImportError:
    from semver import VersionInfo

import verilogae

from DMT.core import unit_registry, VAFile
from DMT.core.mc_parameter import McParameterCollection, McParameter


unit_converter = {
    "s": unit_registry.second,
    "sec": unit_registry.second,
    "A": unit_registry.ampere,
    "A^2s": unit_registry.ampere * unit_registry.ampere * unit_registry.second,
    "V": unit_registry.volt,
    "1/V": 1 / unit_registry.volt,
    "K": unit_registry.kelvin,
    "1/K": 1 / unit_registry.kelvin,
    "K^-1": 1 / unit_registry.kelvin,
    "C": unit_registry.celsius,
    "ohm": unit_registry.ohm,
    "Ohm": unit_registry.ohm,
    "F": unit_registry.farad,
    "Coul": unit_registry.coulomb,
    "K/W": unit_registry.kelvin / unit_registry.watt,
    "J/W": unit_registry.joule / unit_registry.watt,
    "V/K": unit_registry.volt / unit_registry.kelvin,
    "1/K^2": 1 / unit_registry.kelvin / unit_registry.kelvin,
    "Ws/K": unit_registry.watt * unit_registry.second / unit_registry.kelvin,
    "M^(1-AF)": unit_registry.dimensionless,
    "m": unit_registry.meter,
    "m^2": unit_registry.meter * unit_registry.meter,
    "Am^-1": unit_registry.ampere / unit_registry.meter,
    "Am^-2": unit_registry.ampere / unit_registry.meter / unit_registry.meter,
    "Am^-3": unit_registry.ampere / unit_registry.meter / unit_registry.meter / unit_registry.meter,
    "AV^-3": unit_registry.ampere / unit_registry.volt / unit_registry.volt / unit_registry.volt,
    "AV^-3m": unit_registry.ampere
    / unit_registry.volt
    / unit_registry.volt
    / unit_registry.volt
    * unit_registry.meter,
    "Fm^-1": unit_registry.farad / unit_registry.meter,
    "Fm^-2": unit_registry.farad / unit_registry.meter / unit_registry.meter,
    "cm^-3": 1 / unit_registry.meter / unit_registry.meter / unit_registry.meter,  # mhm centi ?
    "Vm^-1": unit_registry.volt / unit_registry.meter,
    "VA^-1m": unit_registry.volt / unit_registry.ampere * unit_registry.meter,
    "VA^-1m^2": unit_registry.volt
    / unit_registry.ampere
    * unit_registry.meter
    * unit_registry.meter,
    "": unit_registry.dimensionless,
}

SEMVER_MCARD_CURRENT = VersionInfo(major=2, minor=2)


class MCard(McParameterCollection):
    """DMT class that implements attributes and methods that are common between all ModelCards such as HICUM and BSIM.

    Parameters
    ----------
    nodes_list        :  tuple(str)
        Port list for this model.
    default_subckt_name  :  str
        Default name for the subcircuit to be included.
    default_module_name  :  str
        Default name of the module of the VA-File for this device.
    version              :  float
        Version of the model.
    va_file : str, optional
        Path to a Verilog-AMS file
    circuit : {None,:class:`~DMT.core.circuit.Circuit`}
        Circuit to simulate this model card.
    mod_name : str
        Name of the Model, this is used by EvalTradica, so it must fit there, e.g. 'HICUM'
    level_num : str
        Level of the Model, this is used by EvalTradica, so it must fit there, e.g. '2'

    Attributes
    ----------
    nodes_list : tuple(str)
        Port list for this model.
    circuit : {None,:class:`~DMT.core.circuit.Circuit`}
        Circuit to simulate this model card.
    default_subckt_name : str
        Default name for the subcircuit to be included.
    default_module_name : str
        Default name of the module of the VA-File for this device.
    version : float
        Version of the model.
    va_file : str
        Path to a Verilog-AMS file *deprecated*
    va_codes : {os.Pathlike: str}
        Dictionary of relative paths and codes -> full VA code structure.
    ignore_checksum: bool, optional
        If True, the checksum of the save modelcard json is ignored, defaults to False.

    """

    def __init__(
        self,
        nodes_list: list[str],
        default_subckt_name: str,
        default_module_name: str,
        version: Union[str, float] = "-",
        va_file: Optional[Union[str, os.PathLike]] = None,
        va_codes=None,
        vae_module=None,
        directory_va_file: Optional[Union[str, os.PathLike]] = None,
        __MCard__=SEMVER_MCARD_CURRENT,
        ignore_checksum: bool = False,
        pdk_path: str = "",
        pdk_corner: str = "",
        **kwargs,
    ):
        super().__init__(**kwargs)

        if not isinstance(__MCard__, VersionInfo):
            try:
                __MCard__ = VersionInfo.parse(__MCard__)
            except TypeError:
                __MCard__ = VersionInfo.parse(f"{__MCard__:1.1f}.0")

        if __MCard__ == VersionInfo(major=1, minor=0):
            # try to obtain the VA-Code
            if (
                va_file is not None
                and not isinstance(va_file, Path)
                and not Path(va_file).is_file()
            ):
                # in Version 1 the code was not part of the MCard. Try to find it...
                print(
                    "The loaded MCard was from Version 1. The VA-Code was not saved inside of the json-file.  "
                )
                if directory_va_file is not None:
                    print("DMT tries to find it. The search location is " + str(directory_va_file))
                    if not isinstance(directory_va_file, Path):
                        directory_va_file = Path(directory_va_file)

                    if (directory_va_file / va_file).is_file():
                        print("File found in the given folder and will be used!")
                        va_file = directory_va_file / va_file
                    else:
                        print("File not found in the folder. No VA-Code is set by core!")
                        va_file = None  # delete file name to escape error while try to load
                else:
                    print("No directory for the va-file is set. No VA-Code is set by core!")
                    print(
                        "To set the directory use: MCard.load_json(path_to_json, directory_va_file=path_to_directory)"
                    )
                    va_file = None  # delete file name to escape error while try to load

            if vae_module is not None:
                raise NotImplementedError

        elif __MCard__ == VersionInfo(major=2, minor=0):
            pass  # nothing to do here?!
        elif __MCard__ == VersionInfo(major=2, minor=1):
            pass  # nothing to do here?!
        elif __MCard__ != SEMVER_MCARD_CURRENT:
            raise NotImplementedError("DMT->MCard: Unknown version of MCard to create!")

        self.nodes_list = nodes_list
        self.default_subckt_name = default_subckt_name
        self.default_module_name = default_module_name
        self.version = version
        self.pdk_path = pdk_path
        self.pdk_corner = pdk_corner

        self._va_codes = None
        if va_codes is not None:
            if isinstance(va_codes, dict):
                self._va_codes = VAFile.import_dict(va_codes, ignore_checksum=ignore_checksum)
            else:
                self._va_codes = va_codes
        elif va_file is not None:
            self.set_va_codes(va_file)

        if self._va_codes:
            self.update_from_vae()

    @property
    def va_codes(self) -> Union[VAFile, None]:
        """Return the attribute directly

        Returns
        -------
        VAFile

        """
        return self._va_codes

    # @va_codes.setter
    # def va_codes(self, va_codes_new: VAFile):
    #     """Set the va_codes

    #     Parameters
    #     ----------
    #     va_codes_new : VAFile
    #         New rel paths with codes
    #     """
    #     self._va_codes = va_codes_new

    def set_va_codes(
        self, path_to_main_code: Union[os.PathLike, str], version: Union[str, float] = None
    ):
        """Sets self._va_codes by extracting all included files from the main file down the include tree.

        Parameters
        ----------
        path_to_main_code : Union[os.PathLike, str]
            Relative or Absolute path to the main Verilog-AMS file
        version : Union[str, float], float
            If the given code has a different model version as currently set.

        Raises
        ------
        NotImplementedError
            If a file is included via absolute path.
        """
        if not isinstance(path_to_main_code, Path):
            path_to_main_code = Path(path_to_main_code)

        self._va_codes = VAFile(path_to_main_code.name)
        self._va_codes.read_structure(path_to_main_code.parent)

        if version is not None:
            self.version = version

    def get_verilogae_model(self) -> ModuleType:
        """Returns the Verilogae Model for this modelcard

        Returns
        -------
        ModuleType
            The Veriloae-module

        Raises
        ------
        FileNotFoundError
            Raised if no VA-Code is set.
        """
        if self.va_codes is None:
            raise FileNotFoundError("No VA-Code set for this MCard!")
        return verilogae.load(self.va_codes.root_vfs, vfs=self.va_codes.vfs)  # type: ignore

    def update_from_vae(self, remove_old_parameters=False):
        """
        Updates the modelcard with information such as parameter boundries and default values, nodes, modules and op vars
        obtained from the Verilog-A source code using `VerilogAE <https://dspom.gitlab.io/verilogae/>`__.

        Parameters
        ----------
        remove_old_parameters : bool, optional
            Deletes parameters which are not part of the VA-Code, by default False
        """
        vae_module = self.get_verilogae_model()
        paras_new = []
        # Updated parameter properties
        for para_name, para_properties in vae_module.modelcard.items():
            try:
                # para = next(para for para in self._paras if para.name == para_name)
                para = self.get(para_name)
                self.remove(para_name)
                ty = type(para_properties.default)

                if para.min > para_properties.min:
                    para.min = para_properties.min
                    para.inc_min = para_properties.min_inclusive  # type: ignore

                if para.max < para_properties.max:
                    para.max = para_properties.max
                    para.inc_max = para_properties.max_inclusive  # type: ignore

                para.unit = unit_converter[para_properties.unit]  # type: ignore
                para.description = para_properties.description  # type: ignore
                para.group = para_properties.group  # type: ignore
                para.val_type = ty  # type: ignore
            except KeyError:
                para = McParameter(
                    para_name,
                    value=para_properties.default,
                    minval=para_properties.min,
                    maxval=para_properties.max,
                    value_type=type(para_properties.default),
                    inc_min=para_properties.min_inclusive,
                    inc_max=para_properties.max_inclusive,
                    exclude=None,  # ? really needed
                    group=para_properties.group,
                    unit=unit_converter[para_properties.unit],
                    description=para_properties.description,
                )
            paras_new.append(para)

        if remove_old_parameters:
            self._paras = []  # remove the old parameters fast...

        for para in paras_new:
            self.add(para, update=False)

        self.update_values()

        self.default_module_name = vae_module.module_name
        self.nodes_list = vae_module.nodes
        # self.version = ?? -> can be set in the submodules according to some model specific stuff

    def read_va_file_boundaries(self, remove_old_parameters=True):
        """Reads the parameter boundaries and possible OPvars from a Verilog-AMS-File

        Needs only to be done once per VA-File, as the boundaries then can
        be saved/loaded from a json (see dump_json)

        Parameters
        ----------
        remove_old_parameters : {True, False}, optional
            If False, parameters which are not found in the VA-File are not removed.
        """
        print(
            "DMT.MCard: Reading from VA-File using this Method is deprecated. Use update_from_vae!"
        )
        paras_new = []

        if self.va_codes is None:
            return

        for va_file, va_code in self.va_codes.iter_codes():
            logging.info(
                "Reading parameter boundaries and operation point variables from %s", va_file
            )

            lines = va_code.code.split("\n")
            group = None
            for line in lines:
                if re.match(r"\s*module", line):
                    # module hic0_full (c,b,e,s,tnode);
                    # module hicumL2va (c,b,e,s,tnode);
                    mo_module_name = re.search(r"module\s(\S+?)\s*\((.+?)\)", line)
                    self.default_module_name = mo_module_name.group(1)
                    self.nodes_list = mo_module_name.group(2).split(",")
                elif line.startswith(r"//"):
                    # only commented lines
                    comment = line.replace("//", "").strip()
                    try:
                        group = self.possible_groups[comment]
                        dummy = 1
                    except KeyError:
                        group = None

                elif re.search(r"parameter", line):
                    # found a parameter line
                    mo_parameter = re.search(
                        r"parameter\s+(real|integer)\s+(\w+)\s*=\s*(\S+)", line
                    )

                    if mo_parameter is None:
                        continue

                    parameter_type = mo_parameter.group(1)
                    parameter_name = mo_parameter.group(2)
                    parameter_default = mo_parameter.group(3)

                    parameter_default = parameter_default.replace(";", "")
                    if parameter_type == "integer":
                        parameter_type = int
                        parameter_default = int(parameter_default)
                    elif parameter_type == "real":
                        parameter_type = float
                        if parameter_default == "`INF":
                            parameter_default = np.inf
                        else:
                            parameter_default = float(parameter_default)
                    else:
                        raise NotImplementedError("Type unknown")

                    tuple_parameter_boundaries = re.findall(r"from\s+(\S+)", line)
                    if tuple_parameter_boundaries:
                        parameter_boundaries = tuple_parameter_boundaries[0].split(":")
                        # at the moment only 1st group is used, if VA-Code with
                        # parameter real a = 1 from 0:10 from 20:30
                        # is analyzed this needs to be extended!

                        if "(" in parameter_boundaries[0]:
                            parameter_include_min = False
                        else:  # else should be "[", not checked now
                            parameter_include_min = True

                        if ")" in parameter_boundaries[1]:
                            parameter_include_max = False
                        else:  # else should be "]", not checked now
                            parameter_include_max = True

                        if parameter_type == int:
                            parameter_min = int(
                                re.search(r"([-.0-9e]+)", parameter_boundaries[0]).group(1)
                            )
                            parameter_max = int(
                                re.search(r"([-.0-9e]+)", parameter_boundaries[1]).group(1)
                            )
                        elif parameter_type == float:
                            try:
                                parameter_min = float(
                                    re.search(r"([-.0-9e]+)", parameter_boundaries[0]).group(1)
                                )
                            except (AttributeError, ValueError):
                                if "INF" in parameter_boundaries[0].upper():
                                    parameter_min = -np.inf
                                else:
                                    raise
                            try:
                                parameter_max = float(
                                    re.search(r"([-.0-9e]+)", parameter_boundaries[1]).group(1)
                                )
                            except (AttributeError, ValueError):
                                if "INF" in parameter_boundaries[1].upper():
                                    parameter_max = np.inf
                                else:
                                    raise
                            # if more are added also add them to the validity checker!!
                        else:  # is impossible, only possibilities are real or integer (see 1st RegExp)
                            raise NotImplementedError(
                                "This parameter type is not implemented: " + str(parameter_type)
                            )
                    else:
                        if parameter_type == float:
                            parameter_min = -np.inf
                            parameter_max = np.inf
                        elif parameter_type == int:  # inf not possible in integer...
                            parameter_min = np.iinfo(int).min
                            parameter_max = np.iinfo(int).max
                        else:
                            raise NotImplementedError()
                        parameter_include_min = True
                        parameter_include_max = True

                    tuple_parameter_excludes = re.findall(r"exclude\s+([0-9:.]+)", line)
                    # at the moment only 1st group is used, if VA-Code with
                    # parameter integer a = 1 from -1:2 exclude 0 exclude 1
                    # parameter integer a = 1 from -1:2 exclude 0:1
                    # is analyzed this needs to be extended!

                    search_unit = re.search(r"unit\s*=\s*\"(.+?)\"", line)
                    if search_unit:
                        # convert from VA to pint...
                        unit = unit_converter[search_unit.group(1)]
                    else:
                        unit = unit_registry.dimensionless

                    search_desc = re.search(
                        r"(desc|info)\s*=\s*\"(.+?)\"", line
                    )  # 'desc' is VA standard, 'info' is ADMS standard
                    if search_desc:
                        desc = search_desc.group(2)
                    else:
                        desc = None

                    try:
                        para = self.get(parameter_name)
                        self.remove(parameter_name)
                    except KeyError:
                        para = McParameter(
                            parameter_name, value=parameter_default, value_type=parameter_type
                        )

                    para.val_type = parameter_type
                    para.max = parameter_max
                    para.min = parameter_min
                    para.inc_max = parameter_include_max
                    para.inc_min = parameter_include_min
                    para.exclude = None
                    #     warnings.warn('The group of the parameter {:s} is unknown from the VA-File'.format(para))
                    # else:
                    if group is not None:
                        para.group = group

                    if desc is not None:
                        para.description = desc

                    if unit is not None:
                        para.unit = unit
                    if tuple_parameter_excludes:  # just assume it is the type parameter...
                        para.exclude = [
                            parameter_type(exclude) for exclude in tuple_parameter_excludes
                        ]

                    paras_new.append(para)
                    # if para in self:
                    #     self.set(para)
                    # else:
                    #     self.add(para)

        if remove_old_parameters:
            self._paras = []  # remove the old parameters fast...

        for para in paras_new:
            self.add(para, update=False)

        self.update_values()

    def info_json(self, save_va_code=True, compress_va_code=False, **kwargs):
        """Returns a dict with serializeable content for the json file to create. Add the info about the concrete subclass to create here!

        Parameters
        ----------
        save_va_code : {True, False}, optional
            If False, the va_codes are not saved...
        compress_va_code : {False, True}, optional
            If True, the codes are saved using zlib compression and a checksum. See:
            https://code.activestate.com/recipes/355486-compress-data-to-printable-ascii-data/
        """
        info_dict = super(MCard, self).info_json(**kwargs)

        info_dict["__MCard__"] = str(
            SEMVER_MCARD_CURRENT
        )  # make versions, so we can introduce compatibility here!

        info_dict["nodes_list"] = self.nodes_list
        info_dict["default_subckt_name"] = self.default_subckt_name
        info_dict["default_module_name"] = self.default_module_name
        info_dict["version"] = self.version
        info_dict["pdk_path"] = self.pdk_path
        info_dict["pdk_corner"] = self.pdk_corner

        if save_va_code and self.va_codes is not None:
            info_dict["va_codes"] = self.va_codes.export_dict(compressed_code=compress_va_code)

        return info_dict

    @classmethod
    def load_json(
        cls,
        file_path: Union[str, Path],
        directory_va_file: Union[str, Path, None] = None,
        ignore_checksum: bool = False,
    ) -> MCard:
        """Load json file

        Just for type hints etc..

        Parameters
        ----------
        file_path : Union[str, Path]
            Path to the json.
        directory_va_file : Union[str, Path, None], optional
            If a relative path to a va_file is set in the modelcard, pass the absolute path to the start folder here, by default None.
            This can be used to load old json modelcards from before saving the full code with the parameters.
        ignore_checksum : bool, optional
            When the code is saved compressed, a checksum is saved with it. If you want to ignore the checksum set this to true, by default False

        Returns
        -------
        MCard
            Loaded modelcard
        """

        return super().load_json(file_path, directory_va_file=directory_va_file, ignore_checksum=ignore_checksum)  # type: ignore

    def get_circuit(self, use_build_in=False, topology=None, **kwargs):
        """Here the modelcard defines it's default simulation circuit.

        Parameters
        ----------
        use_build_in : {False, True}, optional
            Creates a circtui the modelcard using the build-in model
        topology : optional
            If a model has multiple standard circuits, use the topology to differentiate between them..
        """

        raise NotImplementedError("The default modelcard has no circuit :(.")

    def get_build_in(self):
        """Return the parameters embedded in a build-in model (no Va code and correct module name etc)"""
        raise NotImplementedError("The submodels have to implement the build-in parameters")

    def print_to_file(
        self, path_to_file, file_mode="w", subckt_name=None, module_name=None, line_break="\n"
    ):
        """Generates a spectre .lib file which can be included into an netlist.

        Existence of lib file is not checked before writing!
        Name of File: path_to_file + ".lib"

        Parameters
        ----------
        path_to_file : str or os.Pathlike
            Absolute or relative path with file name to the file to generate.
        file_mode : str, optional
            Mode to open the file. Can be used to append :).
        subckt_name : str
            Name of the subcircuit to be included
        module_name : str
            Name of the module from the corresponding VA-File
        line_break : str, optional
            Is added after each parameter, is used as line breaks.
        """
        if subckt_name is None:
            subckt_name = self.default_subckt_name
        if module_name is None:
            module_name = self.default_module_name

        if isinstance(path_to_file, Path):
            path_to_file = path_to_file.with_suffix(".lib")
        else:
            path_to_file = Path(path_to_file + ".lib")

        str_modelcard = "simulator lang = spectre\n"

        str_modelcard += f"subckt {subckt_name:s} (" + " ".join(self.nodes_list) + ")\n"
        str_modelcard += " Q1 (" + " ".join(self.nodes_list) + f") {module_name:s} ({line_break:s}"

        for para in self:
            str_modelcard += f"  {para:<12s} = {para:10.3e} {line_break:s}"

        str_modelcard += ")\n"
        str_modelcard += f"ends {subckt_name:s} \n"

        with path_to_file.open(file_mode) as fp:
            fp.write(str_modelcard)

    def __eq__(self, other):
        """Allows comparing 2 model cards using mc1 == mc2

        mc1 != mc2 is included per default using python3:
        https://docs.python.org/3/reference/datamodel.html#object.__ne__

        """
        if isinstance(other, self.__class__):
            if self.version == other.version:
                # class, version and parameters equal is enough in most cases!
                return self.eq_paras(other)
            else:
                return False

        return NotImplemented

    def load_model(self, path_to_file, force=True):
        """deprecated method, will be removed soon. See load_model_parameters for documentation."""
        warnings.warn(
            "load_model is deprecated and will be renamed in future major releases to load_model_parameters.\n",
            category=DeprecationWarning,
        )
        self.load_model_parameters(path_to_file, force=force)

    def load_model_parameters(self, path_to_file, force=True):
        """Loads the model from a file

        The loading method is determined according to the file ending (last 3 characters!!)
        Possible is "mcp" (see save_model), "txt" or "mat" (planned)

        Parameters
        ----------
        path_to_file : str
            Filename (with ending!) including a relative or absolute path
        force : boolean, optional
            If True, values are force set. Set false if bounds from VA-File are used...
        """
        if not isinstance(path_to_file, Path):
            path_to_file = Path(path_to_file)

        modcard = None

        # Loading protocol depends on file ending
        file_ending = path_to_file.suffix
        if file_ending == ".mcp" or file_ending == ".mcard":
            logging.info("Loading model parameters from pickled model card: %s", str(path_to_file))

            with path_to_file.open("rb") as myfile:
                modcard = pickle.load(myfile)

        elif file_ending == ".mat":
            logging.info("Loading model parameters from mat-File: %s", str(path_to_file))

            modcard = sciio.loadmat(str(path_to_file))
            for key, value in modcard.items():
                if not key.startswith("__"):
                    modcard[key] = np.ndarray.item(value)

        elif file_ending == ".txt":
            logging.info("Loading model parameters from a txt-File: %s", str(path_to_file))

            modcard = []

            str_modelcard = path_to_file.read_text()

            # split it
            re_object = re.findall(r"[a-zA-Z0-9]+\s*=\s*\S+", str_modelcard)

            for param_value in re_object:
                param_value = param_value.split("=")
                modcard.append((param_value.strip(), float(param_value[1].strip())))
        elif file_ending == ".lib" or file_ending == "":
            logging.info("Loading model parameters from a TRADICA lib-File: %s", str(path_to_file))

            modcard = []
            str_lib = path_to_file.read_text()

            # get the model part
            search_parameters = re.search(
                r"(model|subckt)(.*)(ends|)", str_lib, flags=re.DOTALL | re.IGNORECASE
            )
            if search_parameters is not None:
                str_lib = search_parameters.group(2)
            # else there are only parameters

            # split it
            # re_object = re.findall(r"([a-zA-Z0-9]+\s*=\s*[a-zA-Z0-9.+()-]+\s*)", str_lib) # new and better: https://regex101.com/r/Bwvc69/1
            # re_object = re.findall(r"([a-zA-Z0-9]+)\s*=\s*[\(|]\s*(\S+)\s*[\|]", str_lib) # even newer and better https://regex101.com/r/DsZP2J/1
            re_object = re.findall(
                r"([a-zA-Z0-9]+)\s*=\s*((\(|)\s*\S+\s*(\)|))", str_lib
            )  # even newer and better https://regex101.com/r/DsZP2J/2

            for param_name, param_value, _bracket_start, _bracket_close in re_object:
                name = param_name.strip().lower()
                if name == "level" or name == "version" or name == "lang":
                    continue
                value = param_value.strip()
                if (
                    ")" in value and "(" not in value
                ):  # cuts out an probable single closing bracket.
                    value = value.replace(")", "").strip()
                if "***" in value:  # appears sometimes in TRADICA files
                    continue

                value = Calculator.evaluate(value)  # allow calculations in parameter value
                if value is None:
                    raise IOError(
                        "Error while reading lib-file! Could not evaluate the parameter value '"
                        + param_value[1].strip()
                        + "' of the parameter "
                        + name
                        + "!"
                    )
                modcard.append((name, value))
        else:
            raise IOError(
                "Was not able to load parameters from given file!\nGiven was: " + str(path_to_file)
            )

        if modcard is None:
            raise IOError("Loading from file did not work!")

        if isinstance(modcard, list):
            for (parameter_name, parameter_value) in modcard:
                try:
                    # do not reset the limits if parameter is already in modelcard
                    self.set_values({parameter_name: parameter_value}, force=force)
                except KeyError:
                    self.add(McParameter(parameter_name, value=parameter_value))
                except ValueError:
                    # if force==False and parameter value is out of bounds -> do not set the value...
                    warnings.warn(
                        f"DMT->MCard: The parameter {parameter_name} was not loaded from {path_to_file} since the value {parameter_value} was out of bounds.",
                        category=RuntimeWarning,
                    )

                    logging.info(
                        "DMT->MCard: The parameter %s was not loaded from %s since the value %f was out of bounds.",
                        parameter_name,
                        path_to_file,
                        parameter_value,
                    )

        elif isinstance(modcard, McParameterCollection):
            for para in modcard:
                if not hasattr(para, "group"):
                    para.group = None

                try:
                    self.set(para)
                except KeyError:
                    self.add(para)
        elif isinstance(modcard, dict):
            for (parameter_name, parameter_value) in modcard.items():
                if not parameter_name.startswith("__"):
                    try:
                        # do not reset the limits if parameter is already in modelcard
                        self.set_values({parameter_name: parameter_value}, force=force)
                    except KeyError:
                        self.add(McParameter(parameter_name, value=parameter_value))
                    except ValueError:
                        # if force==False and parameter value is out of bounds -> do not set the value...
                        warnings.warn(
                            f"DMT->MCard: The parameter {parameter_name} was not loaded from {path_to_file} since the value {parameter_value} was out of bounds.",
                            category=RuntimeWarning,
                        )

                        logging.info(
                            "DMT->MCard: The parameter %s was not loaded from %s since the value %f was out of bounds.",
                            parameter_name,
                            path_to_file,
                            parameter_value,
                        )
        else:
            raise OSError(
                "Loading from file worked, but I do not know how to handle the loaded content of type "
                + str(type(modcard))
                + "!"
            )


class Calculator(ast.NodeVisitor):
    """Small "safe" calculator to allow calculations of model parameters.

    Avoiding the unsafe "eval"...

    Source:
    https://stackoverflow.com/a/33030616

    """

    _UnaryOP_MAP = {
        ast.UAdd: operator.pos,
        ast.USub: operator.neg,
        ast.Invert: operator.neg,
    }
    """ Implemented unary operators of the calculator (1 Number operations). """
    _BinaryOP_MAP = {
        ast.Add: operator.add,
        ast.Sub: operator.sub,
        ast.Mult: operator.mul,
        ast.Div: operator.truediv,
        ast.Pow: operator.pow,
    }
    """ Implemented binary operators of the calculator (2 Number operations). """

    def visit_UnaryOp(self, node):
        operand = self.visit(node.operand)
        return self._UnaryOP_MAP[type(node.op)](operand)

    def visit_BinOp(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        return self._BinaryOP_MAP[type(node.op)](left, right)

    def visit_Num(self, node):
        return node.n

    def visit_Constant(self, node):
        return node.n

    def visit_Expr(self, node):
        return self.visit(node.value)

    @classmethod
    def evaluate(cls, expression):
        tree = ast.parse(expression)
        calc = cls()
        return calc.visit(tree.body[0])
